from enum import Enum
from typing import Tuple


class Rule(Enum):
    """This enum serves a few purposes:

    1) define symbols for rule IDs to be used in code
    2) map these symbols to names in the UI
    3) collect all rule IDs/provide validation for IDs
    """

    LINE_LENGTH = "line-length"
    TRAILING_WHITESPACE = "trailing-whitespace"
    COMMAND_ARGS = "command-args"
    REDEFINED_BUILTIN = "redefined-builtin"
    UNBRACED_EXPR = "unbraced-expr"
    REDUNDANT_EXPR = "redundant-expr"

    def __str__(self):
        return self.value


ALL_RULES = [rule for rule in Rule]


class Violation:
    def __init__(
        self, id: Rule, message: str, start: Tuple[int, int], end: Tuple[int, int]
    ):
        self.id = id
        self.message = message
        self.start = start
        self.end = end

    def __lt__(self, other):
        return self.start < other.start

    def str(self):
        line, col = self.start
        rule = str(self.id)

        return f"{line}:{col}: {self.message} [{rule}]"

    @classmethod
    def create(cls, id):
        def func(message: str, start: Tuple[int, int], end: Tuple[int, int]):
            return cls(id, message, start, end)

        return func



import codecs
import os
import pathlib
import re
from typing import List, Optional

import pathspec


def register_codec_warning(name):
    def replace_with_warning_handler(e):
        # TODO: formal warning mechanism, include path
        print("Warning: non-unicode characters in file, replacing with �")
        return codecs.replace_errors(e)

    codecs.register_error(name, replace_with_warning_handler)


def resolve_sources(
    paths: List[pathlib.Path],
    exclude_patterns: List[str],
    exclude_root: pathlib.Path,
    extensions: List[str],
) -> List[Optional[pathlib.Path]]:
    """Resolves paths passed via CLI to a list of filepaths to lint.

    `paths` is a list of paths that may be files or directories. Files are
    returned verbatim if they exist, and directories are recursively searched
    for files that have an extension specified in `extensions`. Paths that match a
    pattern in `exclude_patterns` are ignored (based on gitignore pattern
    format, see https://git-scm.com/docs/gitignore#_pattern_format).

    Raises FileNotFoundError if a supplied path does not exist.
    """
    extensions = [f".{ext}" if not ext.startswith(".") else ext for ext in extensions]
    exclude_root = exclude_root.resolve()
    exclude_patterns = [
        re.sub(r"^\s*#", r"\#", pattern) for pattern in exclude_patterns
    ]
    exclude_spec = pathspec.PathSpec.from_lines("gitwildmatch", exclude_patterns)

    def is_excluded(path):
        abspath = path.resolve()
        try:
            relpath = os.path.relpath(abspath, start=exclude_root)
        except ValueError:
            print(
                "Warning: processing files on different drive from where command was"
                " run, 'exclude' config may not behave as expected"
            )
            relpath = abspath

        if exclude_spec.match_file(relpath):
            return True
        return False

    sources: List[Optional[pathlib.Path]] = []

    for path in paths:
        if str(path) == "-":
            sources.append(None)
            continue

        if not path.exists():
            raise FileNotFoundError(f"path {path} does not exist")

        if is_excluded(path):
            continue

        if not path.is_dir():
            sources.append(path)
            continue

        for dirpath, _, filenames in os.walk(path):
            for name in filenames:
                _, ext = os.path.splitext(name)
                if ext.lower() in extensions:
                    child = pathlib.Path(dirpath) / name
                    if not is_excluded(child):
                        sources.append(child)

    return sources



"""End-to-end tests for tclfmt."""

import pathlib
import shutil
import subprocess

MY_DIR = pathlib.Path(__file__).parent.resolve()


def test_tclfmt():
    cmd = ["tclfmt", MY_DIR / "data" / "dirty.tcl"]
    p = subprocess.run(cmd, capture_output=True, cwd=MY_DIR)

    with open(MY_DIR / "data" / "dirty.formatted.tcl", "r") as f:
        expected = f.read()

    assert p.stdout.decode("utf-8") == expected
    assert p.returncode == 0


def test_tclfmt_check():
    cmd = ["tclfmt", "--check", MY_DIR / "data" / "dirty.tcl"]
    p = subprocess.run(cmd, capture_output=True, cwd=MY_DIR)

    assert p.returncode == 1


def test_tclfmt_in_place(tmp_path):
    test_data = MY_DIR / "data" / "dirty.tcl"
    input = tmp_path / "dirty.tcl"
    shutil.copyfile(test_data, input)

    cmd = ["tclfmt", "--in-place", input]
    p = subprocess.run(cmd, capture_output=True, cwd=MY_DIR)

    with open(input, "r") as f:
        actual = f.read()

    with open(MY_DIR / "data" / "dirty.formatted.tcl", "r") as f:
        expected = f.read()

    assert p.returncode == 0
    assert actual == expected



from pathlib import Path

from tclint.main import lint
from tclint.config import Config
from tclint.violations import Rule


def test_cmd_args_in_sub():
    """Ensure command args get checked in cmd sub."""
    script = "[puts]"
    violations = lint(script, Config(), Path())

    assert len(violations) == 1
    assert violations[0].id == Rule("command-args")


def test_no_false_positive_arg_expansion():
    """Ensure that argument count checks don't flag when the number of arguments
    is ambiguous."""
    # $foo may be a list with two items, which is legal
    script = "rename {*}$foo"
    violations = lint(script, Config(), Path())
    assert len(violations) == 0


def test_accidental_return_expr():
    script = "return 5 + 2"
    violations = lint(script, Config(), Path())
    assert len(violations) == 1
    assert violations[0].id == Rule("command-args")


def test_ignore_path():
    script = "expr $foo"
    fake_path = Path("bad.tcl")
    violations = lint(script, Config(), fake_path)
    assert len(violations) == 1

    config = Config(ignore=[{"path": fake_path, "rules": [Rule("unbraced-expr")]}])
    violations = lint(script, config, fake_path)
    assert len(violations) == 0


def test_redefined_builtin():
    script = "proc puts {} {}"
    violations = lint(script, Config(), Path())
    assert len(violations) == 1
    assert violations[0].id == Rule("redefined-builtin")
    assert violations[0].start == (1, 1)
    assert violations[0].end == (1, 13)


def test_no_violation_multiline_expr():
    script = r"""
if {1 && \
    2} {}"""
    violations = lint(script, Config(), Path())
    assert len(violations) == 0


def test_unbraced_expr_varsub():
    script = r"expr $foo"
    violations = lint(script, Config(), Path())
    assert len(violations) == 1
    assert violations[0].id == Rule.UNBRACED_EXPR
    assert violations[0].start == (1, 6)
    assert violations[0].end == (1, 10)


def test_unbraced_expr_with_braced_word():
    script = r"expr 1 + {2}"
    violations = lint(script, Config(), Path())
    assert len(violations) == 1
    assert violations[0].id == Rule.UNBRACED_EXPR
    assert violations[0].start == (1, 6)
    assert violations[0].end == (1, 13)

    script = r'expr 1 + "2"'
    violations = lint(script, Config(), Path())
    assert len(violations) == 1
    assert violations[0].id == Rule.UNBRACED_EXPR
    assert violations[0].start == (1, 6)
    assert violations[0].end == (1, 13)

    script = r"expr {1} + {2}"
    violations = lint(script, Config(), Path())
    assert len(violations) == 1
    assert violations[0].id == Rule.UNBRACED_EXPR
    assert violations[0].start == (1, 6)
    assert violations[0].end == (1, 15)


def test_redundant_expr_check():
    script = r"""
expr {[expr 1]}
expr {([expr 1] + [expr 2])}
expr {[expr 1] ? [expr 2] : [expr 3]}
expr {max([expr 1], [expr 2])}
""".strip()

    violations = lint(script, Config(), Path())
    assert len(violations) == 8
    assert all(v.id == Rule.REDUNDANT_EXPR for v in violations)

    assert violations[0].start == (1, 7)
    assert violations[0].end == (1, 15)
    assert violations[1].start == (2, 8)
    assert violations[1].end == (2, 16)
    assert violations[2].start == (2, 19)
    assert violations[2].end == (2, 27)
    assert violations[3].start == (3, 7)
    assert violations[3].end == (3, 15)
    assert violations[4].start == (3, 18)
    assert violations[4].end == (3, 26)
    assert violations[5].start == (3, 29)
    assert violations[5].end == (3, 37)
    assert violations[6].start == (4, 11)
    assert violations[6].end == (4, 19)
    assert violations[7].start == (4, 21)
    assert violations[7].end == (4, 29)


def test_proc_args_bad():
    script = r"proc foo { { i 1 2 } } { }"
    violations = lint(script, Config(), Path())
    assert violations[0].id == Rule("command-args")


def test_expr_no_args():
    script = "expr"
    violations = lint(script, Config(), Path())
    assert len(violations) == 1
    assert violations[0].id == Rule("command-args")


def test_proc_no_args():
    script = "proc"
    violations = lint(script, Config(), Path())
    assert len(violations) == 1
    assert violations[0].id == Rule("command-args")
